#include <cmath>
#include <fstream>
#include <iostream>
#include <limits>

// the size of the input radiance image
const int xPixelCount = 320;
const int yPixelCount = 240;

/**
 * A simple enum to indicate if we are on a "red" or "blue" row
 */
enum RowType {
    RedRow,
    BlueRow
};

/**
 * A simple enum to track the filter type
 */
enum FilterType {
    RedFilter   = 0,
    GreenFilter = 1,
    BlueFilter  = 2
};

/**
 * This little function extracts the value for given pixel coordinate.
 * If the coordinate is outside the image bounds, we return 0.
 */
double getValue( double * _image, int _xx, int _yy )
{
    double pixelValue = 0.0;
    if(( _xx >= 0 ) && ( _xx < xPixelCount )
    && ( _yy >= 0 ) && ( _yy < yPixelCount )) {
        pixelValue = _image[( _yy * xPixelCount ) + _xx ];
    }
    return pixelValue;
}

int
main( int argCount, char ** argVector )
{
    // open the input radiance file and read it into a 1D array
    double * rawPixels = new double[ xPixelCount * yPixelCount ];
    std::ifstream inputFile( "bayer.img" );
    inputFile.read(( char * )rawPixels,
            xPixelCount * yPixelCount * sizeof( double ));
    if( inputFile.fail() ) {
        std::cerr
            << "debayer: Could not open and read the input file!"
            << std::endl;
        return 1;
    }

    // figure out the min/max of all the pixels so we can scale to 8-bits
    double minValue = +std::numeric_limits< double >::max();
    double maxValue = -std::numeric_limits< double >::max();
    for( int yy = 0; yy < yPixelCount; yy++ ) {
        for( int xx = 0; xx < xPixelCount; xx++ ) {
            double pixelValue = getValue( rawPixels, xx, yy );
            if( minValue > pixelValue ) {
                minValue = pixelValue;
            }
            if( maxValue < pixelValue ) {
                maxValue = pixelValue;
            }
        }
    }

    // compute the gain and bias for a linear scaling
    double gain = 255.0 / ( maxValue - minValue );
    double bias = -minValue * gain;

    // open the output PNM file and write the header
    std::ofstream outputFile( "debayer.pnm" );
    outputFile
        << "P3" << std::endl
        << xPixelCount << " " << yPixelCount << std::endl
        << 255 << std::endl;

    // iterate through all the Y rows in this image
    for( int yy = 0; yy < yPixelCount; yy++ ) {

        // figure out if this "red" or "blue" row
        RowType rowType;
        if(( yy % 2 ) == 0 ) {
            rowType = RedRow;
        }
        else {
            rowType = BlueRow;
        }

        // iterate through all the X pixels in this row
        for( int xx = 0; xx < xPixelCount; xx++ ) {

            // figure out what filter it had
            FilterType filterType;
            if( rowType == RedRow ) {
                if(( xx % 2 ) == 0 ) {
                    filterType = RedFilter;
                }
                else {
                    filterType = GreenFilter;
                }
            }
            else if( rowType == BlueRow ) {
                if(( xx % 2 ) == 1 ) {
                    filterType = BlueFilter;
                }
                else {
                    filterType = GreenFilter;
                }
            }

            // figure out how to get the other colors
            double redValue, greenValue, blueValue;
            if( filterType == RedFilter ) {
                redValue = getValue( rawPixels, xx, yy );
                greenValue =
                    ( getValue( rawPixels, xx-1, yy )
                        + getValue( rawPixels, xx+1, yy )
                        + getValue( rawPixels, xx, yy-1 )
                        + getValue( rawPixels, xx, yy+1 ) ) / 4.0;
                blueValue =
                    ( getValue( rawPixels, xx-1, yy-1 )
                        + getValue( rawPixels, xx+1, yy-1 )
                        + getValue( rawPixels, xx+1, yy+1 )
                        + getValue( rawPixels, xx-1, yy+1 ) ) / 4.0;
            }
            else if( filterType == GreenFilter ) {
                greenValue = getValue( rawPixels, xx, yy );
                if( rowType == BlueRow ) {
                    redValue =
                        ( getValue( rawPixels, xx, yy-1 )
                            + getValue( rawPixels, xx, yy+1 ) ) / 2.0;
                    blueValue =
                        ( getValue( rawPixels, xx-1, yy )
                            + getValue( rawPixels, xx+1, yy ) ) / 2.0;
                }
                else {
                    redValue =
                        ( getValue( rawPixels, xx-1, yy )
                            + getValue( rawPixels, xx+1, yy ) ) / 2.0;
                    blueValue =
                        ( getValue( rawPixels, xx, yy-1 )
                            + getValue( rawPixels, xx, yy+1 ) ) / 2.0;
                }
            }
            else if( filterType == BlueFilter ) {
                redValue =
                    ( getValue( rawPixels, xx-1, yy-1 )
                        + getValue( rawPixels, xx+1, yy-1 )
                        + getValue( rawPixels, xx+1, yy+1 )
                        + getValue( rawPixels, xx-1, yy+1 ) ) / 4.0;
                greenValue =
                    ( getValue( rawPixels, xx-1, yy )
                        + getValue( rawPixels, xx+1, yy )
                        + getValue( rawPixels, xx, yy-1 )
                        + getValue( rawPixels, xx, yy+1 ) ) / 4.0;
                blueValue = getValue( rawPixels, xx, yy );
            }

            // scale the values
            redValue = ( redValue * gain ) + bias;
            greenValue = ( greenValue * gain ) + bias;
            blueValue = ( blueValue * gain ) + bias;

            // threshold the values
            if( redValue > 255.0 ) {
                redValue = 255.0;
            }
            if( greenValue > 255.0 ) {
                greenValue = 255.0;
            }
            if( blueValue > 255.0 ) {
                blueValue = 255.0;
            }

            // write the interpolated pixel values out to the PNM file
            outputFile
                << static_cast< int >( redValue ) << " "
                << static_cast< int >( greenValue ) << " "
                << static_cast< int >( blueValue ) << std::endl;
        }
    }
}
